/**
 * Example program for controller the LEDs on Atari controllers from an SDL
 * program.
 *
 * This assumes libSDL2 has been patched with an Atari specific API to
 * translate an SDL_Joystick into a linux device name for use with the
 * atari led API. For those building their own SDL the relevant patch can be
 * found in this directory as Expose-Linux-device-path-for-Joysticks.patch.
 *
 * This program can be build using the following command:
 *   gcc sdl-controller-leds.c -o sdl-controller-leds  \
 *     $(pkg-config --cflags * --libs sdl2 atari-leds)
 *
 * When running the program the brightness of each of the 4 LEDS on the modern
 * controller will be synced to two analog sticks, while for a classic
 * controller a single led with be lit following the twist of the stick
 */
#include <stdio.h>
#include <stdlib.h>

#include <atari-leds.h>
#include <SDL.h>

typedef struct {
  SDL_Joystick *sdl;
  struct atari_leds *leds;
} controller;

#define N_CONTROLLERS 4
controller controllers[N_CONTROLLERS] = {
  { NULL, NULL },
};

void
update_controller (controller *c)
{
  if (c->sdl == NULL)
    return;

  atari_leds_clear (c->leds);
  if (atari_leds_get_device_type (c->leds) == ATARI_LEDS_MODERN) {
    int i;
    /* Match the led brightness of the 4 leds to the axis of the 4 analog stick
     * axis  */
    for (i = 0; i < 4 ; i++) {
      Sint32 v;
      uint8_t l;

      /* Axis 0 and 1 are left stick, 3 and 4 right stick */
      v = SDL_JoystickGetAxis (c->sdl, i >= 2 ? i + 1 : i);
      v = SDL_abs(v);

      l = SDL_min(SDL_MAX_UINT8, (SDL_MAX_UINT8 * v) / SDL_MAX_SINT16);

      atari_leds_set_led (c->leds, i, l);
    }
  } else { /* classic */
      /* Light one led, following the joystick twist */
      Sint32 v = SDL_JoystickGetAxis (c->sdl, 0);
      int led;
      v += SDL_MAX_SINT16;

      led = (atari_leds_get_n_leds (c->leds) * v) / (2 * SDL_MAX_SINT16);
      atari_leds_set_led (c->leds, led, SDL_MAX_UINT8);
  }
  atari_leds_commit (c->leds);
}

SDL_bool
maybe_add_controller (int device)
{
  struct atari_leds *leds;
  const char *name = SDL_JoystickNameForIndex(device);
  Uint16 vendor = SDL_JoystickGetDeviceVendor (device);
  Uint16 product = SDL_JoystickGetDeviceProduct (device);
  SDL_Joystick *j;
  int i;

  SDL_Log("Found: %s - %04x:%04x (%d)\n", name, vendor, product, device);
  if (vendor != 0x3250) {
    SDL_Log("Not an Atari controller\n");
    return SDL_FALSE;
  }

  j = SDL_JoystickOpen (device);
  if (j == NULL) {
    SDL_Log("Failed to open controller\n");
    return SDL_FALSE;
  }

  leds = atari_leds_new_from_input (VCS_SDL_JoystickLinuxDevicePath (j));
  if (leds == NULL) {
    SDL_Log("Failed create leds device for controller\n");
    return SDL_FALSE;
  }

  for (i = 0 ; i < N_CONTROLLERS; i++) {
    controller *c = &controllers[i];
    if (c->sdl == NULL) {
      c->sdl = j;
      c->leds = leds;
      SDL_Log("Added as instance %d\n", SDL_JoystickInstanceID (c->sdl));
      return SDL_TRUE;
    }
  }

  SDL_Log("Only supported up to %d controllers", N_CONTROLLERS);
  SDL_JoystickClose (j);
  atari_leds_free (leds);
  return SDL_FALSE;
}

void
maybe_remove_controller (int device)
{
  int i;

  for (i = 0 ; i < N_CONTROLLERS; i++) {
    controller *c = &controllers[i];
    if (c->sdl == NULL)
      continue;

    if (SDL_JoystickInstanceID (c->sdl) == device) {
      SDL_Log("Removed instance: %d\n", device);
      SDL_JoystickClose (c->sdl);
      c->sdl = NULL;

      atari_leds_free (c->leds);
      c->leds = NULL;
    }
  }
}

int
main (int argc, char **argv)
{
  int quit = 0;
  int i;

  if (SDL_Init (SDL_INIT_JOYSTICK) < 0) {
    SDL_LogError(SDL_LOG_CATEGORY_APPLICATION, "Failed to init SDL");
    exit (1);
  }

  /* Main loop; When events occur pop the whole event queue, then update
   * controller leds to match the current state and sleep for a bit to get
   * a maximum of 30 updates per second when things change 
   */
  while (!quit) {
    SDL_Event event;

    SDL_Delay (33);
    SDL_WaitEvent (&event);
    do {
      if (event.type == SDL_QUIT) {
        quit = 1;
      } else if (event.type == SDL_JOYDEVICEADDED) {
        maybe_add_controller (event.jdevice.which);
      } else if (event.type == SDL_JOYDEVICEREMOVED) {
        maybe_remove_controller (event.jdevice.which);
      }
    } while (SDL_PollEvent (&event) > 0);

    for (i = 0; i < N_CONTROLLERS; i++) {
        update_controller (&controllers[i]);
    }
  }

  for (i = 0; i < N_CONTROLLERS; i++) {
    if (controllers[i].leds != NULL)
      atari_leds_drop_control (controllers[i].leds);
  }
  return 0;
}
