#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <unistd.h>
#include <getopt.h>

#include <signal.h>

#include <math.h>

#include <string.h>

#include <atari-leds.h>

static bool should_stop = false;

typedef void pattern (struct atari_leds *leds, uint count, int hz, int cycle);

static void
sine (struct atari_leds *leds, uint count, int hz, int cycle) {
  double x;

  x = (count % cycle) * 2 * M_PI / cycle;

  atari_leds_set_all (leds, 127 * sin(x) + 128);
}

static void
dot (struct atari_leds *leds, uint count, int hz, int cycle) {
  int n = atari_leds_get_n_leds (leds);

  atari_leds_clear (leds);
  atari_leds_set_led (leds,
    (count * n / cycle)  % n, 255);
}

static void
block (struct atari_leds *leds, uint count, int hz, int cycle) {
  int n = atari_leds_get_n_leds (leds);
  unsigned int blocksize =  n/8 + 1;

  atari_leds_clear (leds);
  atari_leds_set_block (leds, (count * n / cycle) % n, blocksize, 255);
}

static void
swipe (struct atari_leds *leds, uint count, int hz, int cycle) {
  unsigned int offset;
  unsigned int o;
  unsigned int blocksize;
  unsigned int max;
  int n = atari_leds_get_n_leds (leds);

  blocksize =  n/8 + 1;
  max = n - blocksize;

  o =  (count * max * 2/ cycle) % (max * 2);
  offset = o > max ? max * 2 - o : o;

  atari_leds_clear (leds);
  atari_leds_set_block (leds, offset, blocksize, 255);
}

static void
wiper (struct atari_leds *leds, uint count, int hz, int cycle) {
  unsigned int offset;
  unsigned int o;
  int n = atari_leds_get_n_leds (leds);

  o =  (count * n * 2/ cycle) % (n * 2);
  offset = o > n ? n * 2 - o : o;

  atari_leds_clear (leds);
  atari_leds_set_block (leds, 0, offset, 255);
}

static void
blink (struct atari_leds *leds, uint count, int hz, int cycle) {
  /* Blink at half the cycle count */
  if (count % (cycle) == 0)
    atari_leds_clear (leds);
  else if (count % (cycle/2) == 0) {
    atari_leds_set_all (leds, 255);
  }
}

static void all (struct atari_leds *leds, uint count, int hz, int cycle);

static struct {
    pattern *pattern;
    char *name;
    char *description;
} patterns[] =  {
  { all, "all", "Play all patterns one by one" },
  { blink, "blink", "blink all leds" },
  { dot, "dot", "One dot cycling around" },
  { block, "block", "A block cycling around" },
  { swipe, "swipe", "A block going back and forward" },
  { wiper, "wiper", "Wiping motion" },
  { sine, "sine", "Led brightness following the sine wave" },
  { NULL, NULL }
};

static void
all (struct atari_leds *leds, uint count, int hz, int cycle) {
  int n_p = sizeof(patterns)/sizeof(patterns[0]);
  static int prev_i = 0;
  int i = count/(cycle * 2) % (n_p - 2) + 1;

  if (prev_i != i) {
    printf ("Playing pattern %s\n", patterns[i].name);
    prev_i = i;
  }
  patterns[i].pattern(leds, count, hz, cycle);
}

void usage (char *prog) {
  int i = 0;

  printf("Usage: %s [OPTIONS] <input device file>\n", prog);
  printf("\t-f, --frequency <frequency>: Update frequency\n");
  printf("\t-c, --cycle <cycle count>: Number of updates for a full animation cycle\n");
  printf("\t-p, --pattern <pattern>: Play a specific pattern\n");
  printf("\t-h, --help: Show this help text\n");
  printf("\t\tAvailable patterns:\n");
  for  (i = 0 ; patterns[i].name != NULL; i++) {
    printf("\t\t- %s: %s\n", patterns[i].name, patterns[i].description);
  }
}

void
handle_sigint (int signal) {
  should_stop = true;
}

int
main (int argc, char **argv)
{
  struct atari_leds *leds;
  static struct option options[] = {
    { "frequency",  required_argument, 0, 0 },
    { "cycle",      required_argument, 0, 0 },
    { "help",       no_argument,       0, 0 },
    { "pattern",    required_argument, 0, 0 },
    { 0,            0,                 0, 0 }
  };
  int hz = 30;
  int cycle = -1;
  char c;
  uint count = 0;
  pattern *p = all;

  while ((c = getopt_long (argc, argv, "c:f:hp:", options, NULL)) != -1) {
    char *endp;
    switch (c) {
      case 'c':
        cycle = strtol (optarg, &endp, 10);
        if (*endp != '\0') {
          fprintf (stderr, "Failed to parse: %s\n\n", optarg);
          usage(argv[0]);
          exit(1);
        }
        break;
      case 'f':
        hz = strtol (optarg, &endp, 10);
        if (*endp != '\0') {
          fprintf (stderr, "Failed to parse: %s\n\n", optarg);
          usage(argv[0]);
          exit(1);
        }
        break;
      case 'h':
        usage(argv[0]);
        exit(0);
        break;
      case 'p': {
        int i;
        p = NULL;
        for (i = 0 ; patterns[i].name != NULL; i++) {
          if (!strcmp (patterns[i].name, optarg)) {
            p = patterns[i].pattern;
          }
        }
        if (p == NULL) {
          fprintf (stderr, "Unknown pattern: %s\n\n", optarg);
          exit (1);
        }
        break;
      }
      default:
        fprintf(stderr, "Incorrect options\n\n");
        usage(argv[0]);
        exit(1);
        break;
    }
  }

  if (argc - optind != 1) {
    fprintf (stderr, "Missing input device argument\n\n");
    usage (argv[0]);
    exit (1);
    return 1;
  }

  signal (SIGINT, handle_sigint);

  leds = atari_leds_new_from_input (argv[optind]);
  if (leds == NULL) {
    fprintf (stderr, "Failed to open device\n");
    exit(1);
  }

  printf ("Device opened; Type: %s controller\n",
    atari_leds_get_device_type (leds) == ATARI_LEDS_CLASSIC
      ? "classic": "modern");

  if (cycle == -1)
    cycle = hz * 2;

  while (should_stop == false) {
    p (leds, count++, hz, cycle);
    atari_leds_commit (leds);
    usleep (1000000/hz);
  }

  /* Let the leds go back to firmare control */
  atari_leds_drop_control (leds);
  atari_leds_free (leds);
  return 0;
}
