#ifndef _GNU_SOURCE
#define _GNU_SOURCE 1
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/sysmacros.h>
#include <ctype.h>

#include <stdbool.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <libudev.h>
#include "atari-leds.h"

#define MODERN_LEDS 4
#define CLASSIC_LEDS 24

struct led {
  /* Path relative to the dirfd */
  char *ledpath;
};

struct atari_leds {
  AtariDeviceType dtype;

  int dirfd;

  int n_leds;
  struct led *leds;
  /* current brightness level */
  uint8_t *current;
  /* pending brightness level */
  uint8_t *pending;

  bool firmware_control;
};

static int
_led_get_main_nr (struct udev_device *led)
{
  const char *name = udev_device_get_sysname (led);
  char *nr = strrchr (name, ':');
  char *color;

  if (nr == NULL)
    return -1;
  nr++;

  /* expected 2 digits for the nr */
  if (strlen (nr) != 2)
    return -1;

  if (!isdigit (*nr) || !isdigit (*(nr + 1)))
    return -1;

  for (color = nr - 2; color > name; color -= 1)
    {
      if (*color == ':') {
        color++;
        /* main leds are amber */
        if (strncmp (color, "amber", nr - color - 1 ))
          return -1;

        return atoi (nr);
      }
    }

  return -1;
}

static void
_setup_leds (struct atari_leds *leds, struct udev *u, struct udev_device *hid)
{
  struct udev_enumerate *e = NULL;
  struct udev_list_entry *entry = NULL;
  const char *base;
  size_t base_len;

  e = udev_enumerate_new (u);
  udev_enumerate_add_match_parent (e, hid);
  udev_enumerate_add_match_subsystem (e, "leds");
  udev_enumerate_scan_devices (e);

  base = udev_device_get_syspath (hid);
  base_len = strlen (base);

  udev_list_entry_foreach (entry, udev_enumerate_get_list_entry (e)) {
    struct udev_device *led;
    int lednr;

    led = udev_device_new_from_syspath (u, udev_list_entry_get_name (entry));
    lednr = _led_get_main_nr (led);

    if (lednr >= 0
        && lednr < leds->n_leds
        && leds->leds[lednr].ledpath == NULL) {
      const char *ledpath = udev_device_get_syspath (led);

      if (!strncmp(ledpath, base, base_len)) {
        leds->leds[lednr].ledpath = strdup (ledpath + base_len + 1);
      }
    }

    udev_device_unref (led);
  }

  udev_enumerate_unref (e);
}

static AtariDeviceType
_determine_device_type (struct udev *u, struct udev_device *dev)
{
  struct udev_device *input;
  const char *product;
  AtariDeviceType t = ATARI_LEDS_UNKNOWN;
  char *vid = NULL;
  char *pid = NULL;
  char *pstart;
  char *pend;

  input = udev_device_get_parent_with_subsystem_devtype (dev,
    "input", NULL);

  if (input == NULL)
    return ATARI_LEDS_UNKNOWN;

  /* product string is bus/vid/pid/version */
  product = udev_device_get_property_value (input, "PRODUCT");

  pstart = strchr (product, '/');
  if (pstart == NULL)
    goto out;

  pstart++;

  pend = strchr (pstart, '/');
  if (pend == NULL)
    goto out;

  vid = strndup (pstart, pend - pstart);

  pstart = pend + 1;
  pend = strchr (pstart, '/');
  if (pend == NULL)
    goto out;

  pid = strndup (pstart, pend - pstart);

  if (vid && pid) {
    if (strcmp("3250", vid) == 0) {
      if (strcmp("1001", pid) == 0)
        t = ATARI_LEDS_CLASSIC;
      else if (strcmp("1002", pid) == 0)
        t = ATARI_LEDS_MODERN;
    }
  }

out:
  free(pid);
  free(vid);

  return t;
}

void
atari_leds_free (struct atari_leds *leds)
{
  int i;

  if (leds == NULL)
    return;

  if (leds->dirfd >= 0)
    atari_leds_drop_control (leds);

  if (leds->leds)
    for (i =0 ; i < leds->n_leds; i++)
      free (leds->leds[i].ledpath);

  free (leds->leds);
  free (leds->current);
  free (leds->pending);

  if (leds->dirfd >= 0)
    close (leds->dirfd);

  free (leds);
}

struct atari_leds *
atari_leds_new_from_input (const char *devpath)
{
  struct atari_leds *r = NULL;
  struct udev *u = NULL;
  struct udev_device *dev = NULL, *hid = NULL;
  struct led_entries;
  struct stat sbuf;
  AtariDeviceType dtype;

  u = udev_new();
  if (u == NULL)
    goto fail;

  if (stat (devpath, &sbuf) != 0)
    goto fail;

  dev = udev_device_new_from_devnum (u, 'c', sbuf.st_rdev);
  if (dev == NULL)
    goto fail;

  dtype = _determine_device_type (u, dev);
  if (dtype == ATARI_LEDS_UNKNOWN) {
    goto fail;
  }

  hid = udev_device_get_parent_with_subsystem_devtype (dev,
    "hid", NULL);
  if (hid == NULL)
    goto fail;

  r = malloc(sizeof(*r));
  if (r == NULL)
    goto fail;

  r->dtype = dtype;
  if (dtype == ATARI_LEDS_CLASSIC)
    r->n_leds = CLASSIC_LEDS;
  else
    r->n_leds = MODERN_LEDS;

  r->leds = calloc (r->n_leds, sizeof(struct led));
  if (r->leds == NULL)
    goto fail;

  r->current = calloc (r->n_leds, sizeof(uint8_t));
  if (r->current == NULL)
    goto fail;

  r->pending = calloc (r->n_leds, sizeof(uint8_t));
  if (r->pending == NULL)
    goto fail;

  /* Assume leds are under firmware control initially */
  r->firmware_control = true;

  _setup_leds (r, u, hid);

  r->dirfd = open (udev_device_get_syspath (hid), O_PATH);
  if (r->dirfd < 0)
    goto fail;

out:
  udev_device_unref (dev);
  udev_unref (u);

  return r;
fail:
  atari_leds_free (r);
  r = NULL;

  goto out;
}

int
atari_leds_get_n_leds (struct atari_leds *leds) {
  return leds->n_leds;
}

AtariDeviceType
atari_leds_get_device_type (struct atari_leds *leds) {
  return leds->dtype;
}

static void
_do_set_led (struct atari_leds *leds, unsigned int idx, uint8_t brightness)
{
  struct led *l;
  int dfd, fd;
  char buf[10];
  int len;

  if (leds == NULL || idx >= leds->n_leds)
    return;

  l = &leds->leds[idx];

  len = snprintf(buf, 10, "%d", brightness);
  if (len < 0)
    return;

  dfd = openat (leds->dirfd, l->ledpath, O_PATH);
  if (dfd < 0)
    return;

  fd = openat (dfd, "brightness", O_RDWR);
  close (dfd);
  if (fd < 0) {
    return;
  }

  if (brightness > 255)
    brightness = 255;

  write (fd, buf, len);

  close (fd);
}

void
atari_leds_set_led (struct atari_leds *leds, unsigned int idx, uint8_t brightness)
{
  if (idx >= leds->n_leds)
    return;

  leds->pending[idx] = brightness;
}

void
atari_leds_set_all (struct atari_leds *leds, uint8_t brightness)
{
  memset (leds->pending, brightness, sizeof(uint8_t) * leds->n_leds);
}

void
atari_leds_set_block (struct atari_leds *leds,
                      unsigned int offset,
                      unsigned int number,
                      uint8_t brightness)
{
  if (offset >= leds->n_leds)
    return;

  if (number > leds->n_leds)
    return;

  if (offset + number > leds->n_leds) {
    memset (leds->pending + offset, brightness,
            sizeof(uint8_t) * (leds->n_leds - offset));

    memset (leds->pending, brightness,
            sizeof(uint8_t) * ((offset + number) % leds->n_leds));
  } else {
    memset (leds->pending + offset, brightness, sizeof(uint8_t) * number);
  }
}

void
atari_leds_clear (struct atari_leds *leds)
{
  atari_leds_set_all (leds, 0);
}

void
atari_leds_commit (struct atari_leds *leds)
{
  int i;
  bool fc = leds->firmware_control;

  for (i = 0; i < leds->n_leds; i++) {
    if (fc || leds->pending[i] != leds->current[i])
      _do_set_led (leds, i, leds->pending[i]);
  }
  memcpy (leds->current, leds->pending, sizeof(uint8_t) * leds->n_leds);
  leds->firmware_control = false;
}

void
atari_leds_drop_control (struct atari_leds *leds)
{
  int fd;

  if (leds == NULL)
    return;

  fd = openat (leds->dirfd, "leds_firmware_control", O_RDWR);
  if (fd < 0) {
    return;
  }

  write (fd, "1", 2);
  close (fd);
  leds->firmware_control = true;
}
